"""File algebra.py

:author: Michel Bierlaire
:date: Thu Jun 22 15:09:06 2023 

Function related to linear algebra
"""

import logging
import numpy as np
import scipy.linalg as la
from biogeme_optimization.exceptions import OptimizationError


logger = logging.getLogger(__name__)


def schnabel_eskow(
    the_matrix: np.ndarray,
    tau: float = np.finfo(np.float64).eps ** 0.3333,
    taubar: float = np.finfo(np.float64).eps ** 0.6666,
    mu: float = 0.1,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Modified Cholesky factorization by `Schnabel and Eskow (1999)`_.

    .. _`Schnabel and Eskow (1999)`: https://doi.org/10.1137/s105262349833266x

    If the matrix is 'safely' positive definite, the output is the
    classical Cholesky factor. If not, the diagonal elements are
    inflated in order to make it positive definite. The factor :math:`L`
    is such that :math:`the_matrix + E = PLL^TP^T`, where :math:`E` is a diagonal
    matrix contaninig the terms added to the diagonal, :math:`P` is a
    permutation matrix, and :math:`L` is w lower triangular matrix.

    :param the_matrix: matrix to factorize. Must be square and symmetric.
    :type the_matrix: numpy.array
    :param tau: tolerance factor.
                Default: :math:`\\varepsilon^{\\frac{1}{3}}`.
                See `Schnabel and Eskow (1999)`_
    :type tau: float
    :param taubar: tolerance factor.
                   Default: :math:`\\varepsilon^{\\frac{2}{3}}`.
                   See `Schnabel and Eskow (1999)`_
    :type taubar: float
    :param mu: tolerance factor.
               Default: 0.1.  See `Schnabel and Eskow (1999)`_
    :type mu: float

    :return: tuple :math:`L`, :math:`E`, :math:`P`,
                    where :math:`the_matrix + E = PLL^TP^T`.
    :rtype: numpy.array, numpy.array, numpy.array

    :raises biogeme.exceptions.OptimizationError: if the matrix the_matrix is not square.
    :raises biogeme.exceptions.OptimizationError: if the matrix the_matrix is not symmetric.
    """

    def pivot(j: int) -> None:
        the_matrix[j, j] = np.sqrt(the_matrix[j, j])
        for i in range(j + 1, dim):
            the_matrix[j, i] = the_matrix[i, j] = the_matrix[i, j] / the_matrix[j, j]
            the_matrix[i, j + 1 : i + 1] -= (
                the_matrix[i, j] * the_matrix[j + 1 : i + 1, j]
            )
            the_matrix[j + 1 : i + 1, i] = the_matrix[i, j + 1 : i + 1]

    def permute(i: int, j: int):
        the_matrix[[i, j]] = the_matrix[[j, i]]
        diagonal_matrix[[i, j]] = diagonal_matrix[[j, i]]
        the_matrix[:, [i, j]] = the_matrix[:, [j, i]]
        permutation_matrix[:, [i, j]] = permutation_matrix[:, [j, i]]

    the_matrix = the_matrix.astype(np.float64)
    dim = the_matrix.shape[0]
    if the_matrix.shape[1] != dim:
        raise OptimizationError('The matrix must be square')

    if not np.array_equal(the_matrix, the_matrix.T):
        raise OptimizationError('The matrix must be square and symmetric')

    diagonal_matrix = np.zeros(dim, dtype=np.float64)
    permutation_matrix = np.identity(dim)
    phase_one = True
    gamma = abs(the_matrix.diagonal()).max()
    j = 0
    while j < dim and phase_one is True:
        a_max = np.max(the_matrix.diagonal()[j:])
        a_min = np.min(the_matrix.diagonal()[j:])
        if a_max < taubar * gamma or a_min < -mu * a_max:
            phase_one = False
            break

        # Pivot on maximum diagonal of remaining submatrix
        i = j + np.argmax(the_matrix.diagonal()[j:])
        if i != j:
            # Switch rows and columns of i and j of the_matrix
            permute(i, j)
        if (
            j < dim - 1
            and np.min(
                the_matrix.diagonal()[j + 1 :]
                - the_matrix[j + 1 :, j] ** 2 / the_matrix.diagonal()[j]
            )
            < -mu * gamma
        ):
            phase_one = False  # go to phase two
        else:
            # perform jth iteration of factorization
            pivot(j)
            j += 1

    # Phase two, the_matrix not positive-definite
    if not phase_one:
        if j == dim - 1:
            diagonal_matrix[-1] = delta = -the_matrix[-1, -1] + max(
                tau * (-the_matrix[-1, -1]) / (1 - tau), taubar * gamma
            )
            the_matrix[-1, -1] += delta
            the_matrix[-1, -1] = np.sqrt(the_matrix[-1, -1])
        else:
            delta_prev = 0.0
            gerschgorin = np.zeros(dim)
            k = j - 1  # k = number of iterations performed in phase one
            # Calculate lower Gerschgorin bounds of the_matrix[k+1]
            for i in range(k + 1, dim):
                gerschgorin[i] = (
                    the_matrix[i, i]
                    - np.sum(np.abs(the_matrix[i, k + 1 : i]))
                    - np.sum(np.abs(the_matrix[i + 1 : dim, i]))
                )
            # Modified Cholesky Decomposition
            for j in range(k + 1, dim - 2):
                # Pivot on maximum lower Gerschgorin bound estimate
                i = j + np.argmax(gerschgorin[j:])
                if i != j:
                    # Switch rows and columns of i and j of the_matrix
                    permute(i, j)
                # Calculate E[j, j] and add to diagonal
                norm_j = np.sum(np.abs(the_matrix[j + 1 : dim, j]))
                diagonal_matrix[j] = delta = max(
                    0, -the_matrix[j, j] + max(norm_j, taubar * gamma), delta_prev
                )
                if delta > 0:
                    the_matrix[j, j] += delta
                    delta_prev = delta  # delta_prev will contain E_inf
                # Update Gerschgorin bound estimates
                if the_matrix[j, j] != norm_j:
                    temp = 1.0 - norm_j / the_matrix[j, j]
                    gerschgorin[j + 1 :] += np.abs(the_matrix[j + 1 :, j]) * temp
                # perform jth iteration of factorization
                pivot(j)

            # Final 2 by 2 submatrix
            eigenvalues = np.linalg.eigvalsh(the_matrix[-2:, -2:])
            eigenvalues.sort()
            diagonal_matrix[-2] = diagonal_matrix[-1] = delta = max(
                0,
                -eigenvalues[0]
                + np.maximum(
                    tau * (eigenvalues[1] - eigenvalues[0]) / (1 - tau), taubar * gamma
                ),
                delta_prev,
            )
            if delta > 0:
                the_matrix[-2, -2] += delta
                the_matrix[-1, -1] += delta
                delta_prev = delta
            the_matrix[-2, -2] = np.sqrt(
                the_matrix[-2, -2]
            )  # overwrites the_matrix[-2, -2]
            the_matrix[-1, -2] = (
                the_matrix[-1, -2] / the_matrix[-2, -2]
            )  # overwrites the_matrix[-1, -2]
            the_matrix[-2, -1] = the_matrix[-1, -2]
            # overwrites the_matrix[-1, -1]
            the_matrix[-1, -1] = np.sqrt(
                the_matrix[-1, -1] - the_matrix[-1, -2] * the_matrix[-1, -2]
            )

    lower_triangular_matrix = np.tril(the_matrix)
    return (
        lower_triangular_matrix,
        np.diag(permutation_matrix @ diagonal_matrix),
        permutation_matrix,
    )


def schnabel_eskow_direction(
    gradient: np.ndarray, hessian: np.ndarray, check_convexity: bool = False
) -> np.ndarray:
    """Calculate a descent direction using the Schnabel-Eskow factorization"""
    if check_convexity:
        lower, diagonal, permutation = schnabel_eskow(hessian)
        corrections = diagonal > 0
        if np.any(corrections):
            raise OptimizationError('The quadratic model is not convex.')
    else:
        lower, _, permutation = schnabel_eskow(hessian)
    y3 = -permutation.T @ gradient
    y2 = la.solve_triangular(lower, y3, lower=True, overwrite_b=True)
    y1 = la.solve_triangular(lower.T, y2, lower=False, overwrite_b=True)
    d = permutation @ y1
    return d



"""File neighborhood.py

:author: Michel Bierlaire, EPFL
:date: Wed Jul  5 11:50:56 2023

Abstract class for a neighborhood structure
"""

import logging
import abc
from typing import final, Callable
import numpy as np
from biogeme_optimization.exceptions import OptimizationError
from biogeme_optimization.pareto import SetElement

logger = logging.getLogger(__name__)

OperatorOutput = tuple[SetElement | None, int]
Operator = Callable[[SetElement, int], OperatorOutput]


class OperatorsManagement:
    """
    Class managing the selection and performance analysis of the operators
    """

    def __init__(self, operators: dict[str, Operator]):
        """Ctor

        :param operators: dict where the keys are the names of the
            operators, and the values are the operators themselves. An
            operator is a function that takes two arguments (the
            current solution and the size of the neighborhood), and
            return a neighbor solution.
        :type operators: dict(str: fct)

        """
        self.operators: dict[str, Operator] = operators

        self.scores: dict[str, int] = {k: 0 for k in operators}
        """ dict of scores obtained by the operators"""

        self.names: list[str] = list(operators.keys())
        """ Names of the operators """

        self.available: dict[str, bool] = {k: True for k in operators}
        """ dict of availability of the operators """

        self.last_operator_name: str | None = None

        # This quantity has been calculated to that if one operator
        # has a score of 10, another has a score of -10, and all the
        # others have a score of 1, the operator with the highest
        # score is associated with a probability of about 0.9,
        # irrespectively of the number of operators
        self.scale: float = (
            0.123184457025228 * np.log(float(len(self.scores))) + 0.163017205688887
        )
        self.min_probability: float = 0.1 / len(self.scores)

    def increase_score_last_operator(self) -> None:
        """Increase the score of the last operator.

        :raise OptimizationError: if the operator is not known.
        """
        if self.last_operator_name not in self.scores:
            raise OptimizationError(f'Unknown operator: {self.last_operator_name}')

        self.scores[self.last_operator_name] += 1

    def decrease_score_last_operator(self) -> None:
        """Decrease the score of the last operator. If it has already the minimum
        score, it increases the others.

        :raise OptimizationError: if the operator is not known.
        """
        if self.last_operator_name not in self.scores:
            raise OptimizationError(f'Unknown operator: {self.last_operator_name}')

        self.scores[self.last_operator_name] -= 1

    def probability_from_scores(self) -> np.ndarray | None:
        """Calculates the probability from the scores

        :return: list of probabilities
        :rtype: list(float)

        :raise OptimizationError: if the minimum probability is too large
            for the number of operators. It must be less or equal to 1 /
            N, where N is the number of operators.


        """
        if self.min_probability > 1.0 / len(self.scores):
            raise OptimizationError(
                f'Cannot impose min. probability '
                f'{self.min_probability} '
                f'with {len(self.scores)} operators. '
                f'Maximum is {1.0 / len(self.scores):.3f}.'
            )

        max_score = max(list(self.scores.values()))
        list_of_scores = np.array(
            [
                np.exp(self.scale * (s - max_score)) if self.available[k] else 0
                for k, s in self.scores.items()
            ]
        )
        if len(list_of_scores) == 0:
            return None
        total_score = sum(list_of_scores)
        prob = np.array([float(s) / float(total_score) for s in list_of_scores])
        return self.enforce_minimum_probability(prob, self.min_probability)

    @staticmethod
    def enforce_minimum_probability(
        prob: np.ndarray, min_probability: float
    ) -> np.ndarray:
        """
        :param prob: vector of probabilities
        :type prob: numpy.array

        :param min_probability: minimum probability to select any
                               operator. This is meant to avoid
                               degeneracy, that is to have operators
                               that have no chance to be chosen
        :type min_probability: float

        """
        # Enforce minimum probability
        if not np.isclose(np.sum(prob), 1.0):
            error_msg = (
                f'This is not a valid probability distribution as it '
                f'does not sum up to one but to {np.sum(prob)}: {prob}'
            )
            raise OptimizationError(error_msg)
        if np.any(prob < 0):
            error_msg = (
                f'This is not a valid probability distribution as some '
                f'values are negative : {prob}'
            )
            raise OptimizationError(error_msg)
        ok = prob >= min_probability
        if np.all(ok == False):
            error_msg = (
                f'Impossible to enforce minimum probability {min_probability} '
                f'as all probability are below the threshold: {prob}'
            )
            raise OptimizationError(error_msg)
        too_low = prob < min_probability
        not_zero = prob != 0.0
        update = too_low & not_zero
        reserved_total = update.sum() * min_probability
        if reserved_total >= 1.0:
            error_msg = (
                f'There are two many probabilities below the thresholds. '
                f'Raising them to the threshold would make the sum of all '
                f'probabilities larger than one: '
                f'{update.sum()} * {min_probability} = {reserved_total}.'
            )
            raise OptimizationError(error_msg)

        total_high_scores = prob[ok].sum()
        prob[ok] = (1.0 - reserved_total) * prob[ok] / total_high_scores
        prob[update] = min_probability
        return prob

    def select_operator(self) -> Operator:
        """Select an operator based on the scores

        :return: name of the selected operator
        :rtype: str

        """
        prob = self.probability_from_scores()
        self.last_operator_name = np.random.choice(self.names, 1, p=prob)[0]
        return self.operators[self.last_operator_name]


class Neighborhood(metaclass=abc.ABCMeta):
    """
    Abstract class defining a problem
    """

    def __init__(self, operators: dict[str, Operator]):
        self.operators_management = OperatorsManagement(operators)

    @abc.abstractmethod
    def is_valid(self, element: SetElement) -> tuple[bool, str]:
        """Check the validity of the solution.

        :param element: solution to be checked
        :type element: :class:`biogeme.pareto.SetElement`

        :return: valid, why where valid is True if the solution is
            valid, and False otherwise. why contains an explanation why it
            is invalid.
        :rtype: tuple(bool, str)
        """

    @final
    def generate_neighbor(
        self, element: SetElement, neighborhood_size: int, attempts: int = 5
    ) -> OperatorOutput:
        """Generate a neighbor from the neighborhood of size
        ``neighborhood_size`` using one of the operators

        :param element: current solution
        :type element: SetElement

        :param neighborhood_size: size of the neighborhood
        :type neighborhood_size: int

        :param attempts: number of attempts until we give up
        :type attempts: int

        :return: the neighbor, and the number of modifications actually made

        """
        # Select one operator.
        for _ in range(attempts):
            operator = self.operators_management.select_operator()
            neighbor, number_of_changes = operator(element, neighborhood_size)
            if number_of_changes > 0:
                return neighbor, number_of_changes
        return element, 0

    def last_neighbor_rejected(self) -> None:
        """Notify that a neighbor has been rejected by the
        algorithm. Used to update the statistics on the operators.

        """
        self.operators_management.decrease_score_last_operator()

    def last_neighbor_accepted(self) -> None:
        """Notify that a neighbor has been accepted by the
        algorithm. Used to update the statistics on the operators.

        """
        self.operators_management.increase_score_last_operator()



"""File hybrid_function.py

:author: Michel Bierlaire
:date: Sun Jul  2 15:34:16 2023

Class for the hybrid function calculation, that mixed analytical hessian and BFGS updates
"""

import logging
import numpy as np

from biogeme_optimization.bounds import Bounds
from biogeme_optimization.function import FunctionData, FunctionToMinimize
from biogeme_optimization.bfgs import bfgs

logger: logging.Logger = logging.getLogger(__name__)


class HybridFunction:
    """Object in charge of evaluating the function and its derivative,
    where the second derivative matrix can be either calculated
    analytically or using BFGS updates.

    """

    def __init__(
        self,
        the_function: FunctionToMinimize,
        proportion_analytical_hessian: float,
        bounds: Bounds,
    ) -> None:
        """Constructor

        :param the_function: object to calculate the objective
            function and its derivatives.
        :type the_function: optimization.FunctionToMinimize

        :param proportion_analytical_hessian: proportion of the iterations where
                                  the true hessian is calculated. When
                                  not, the BFGS update is used. If
                                  1.0, it is used for all
                                  iterations. If 0.0, it is not used
                                  at all.
        :type proportion_analytical_hessian: float

        :param bounds: object describing the bound constraints. Used
            to project the gradient and check optimality.
        :type bounds: Bounds

        """
        self.the_function: FunctionToMinimize = the_function
        self.proportion: float = proportion_analytical_hessian
        self.bounds: Bounds = bounds
        self.number_of_analytical_hessians: int = 0
        self.number_of_matrices: int = 0
        self.number_of_numerical_issues: int = 0
        self.previous_x: np.ndarray | None = None
        self.previous_gradient: np.ndarray | None = None
        self.previous_hessian: np.ndarray | None = None

    def can_calculate_analytical(self) -> bool:
        """Determines if the analytical hessian can be calculated or not"""

        if self.proportion == 0:
            return False
        if self.proportion == 1:
            return True
        if self.number_of_matrices == 0:
            return True
        return (
            float(self.number_of_analytical_hessians) / float(self.number_of_matrices)
            <= self.proportion
        )

    def calculate_function(self, iterate: np.ndarray) -> float:
        """Calculates the canonical_value of the function

        :param iterate: values
        :type iterate: numpy.array
        """
        self.the_function.set_variables(iterate)
        return self.the_function.f()

    def calculate_function_and_derivatives(
        self, iterate: np.ndarray
    ) -> FunctionData | None:
        """Calculates the function, its gradient, and the hessian, or its approximation

        :param iterate: values
        :type iterate: numpy.array
        """
        if self.can_calculate_analytical():
            function_data = self._calculate_function_and_derivatives_analytical(iterate)
        else:
            function_data = self._calculate_function_and_derivatives_bfgs(iterate)
        if function_data is None:
            return None
        self.previous_x = iterate
        self.previous_gradient = function_data.gradient
        self.previous_hessian = function_data.hessian
        return function_data

    def _calculate_function_and_derivatives_bfgs(
        self, iterate: np.ndarray
    ) -> FunctionData | None:
        """Calculates the function, its gradient, and the BFGS
            approximation of the hessian

        :param iterate: values
        :type iterate: numpy.array

        """
        self.the_function.set_variables(iterate)
        if self.the_function.check_optimality(bounds=self.bounds):
            return None
        if self.previous_x is not None:
            if self.the_function.check_insufficient_progress(self.previous_x):
                return None
        function_data = self.the_function.f_g()
        if self.previous_x is None:
            hessian = np.eye(self.the_function.dimension())
        else:
            delta_x = iterate - self.previous_x
            delta_gradient = function_data.gradient - self.previous_gradient
            hessian = bfgs(self.previous_hessian, delta_x, delta_gradient)

        self.number_of_matrices += 1
        final_function_data = FunctionData(
            function=function_data.function,
            gradient=function_data.gradient,
            hessian=hessian,
        )
        return final_function_data

    def _calculate_function_and_derivatives_analytical(
        self, iterate: np.ndarray
    ) -> FunctionData | None:
        """Calculates the function, its gradient, and the hessian

        :param iterate: values
        :type iterate: numpy.array
        """
        self.the_function.set_variables(iterate)
        if self.the_function.check_optimality(bounds=self.bounds):
            return None
        if self.previous_x is not None:
            if self.the_function.check_insufficient_progress(self.previous_x):
                return None
        function_data = self.the_function.f_g_h()
        self.number_of_analytical_hessians += 1
        self.number_of_matrices += 1

        # If there is a numerical problem with the
        # Hessian, we apply BFGS

        if np.linalg.norm(function_data.hessian) > 1.0e100:
            self.number_of_numerical_issues += 1
            logger.warning(
                f'Numerical problem with the second '
                f'derivative matrix. '
                f'Norm = {np.linalg.norm(function_data.hessian)}. '
                f'Replaced by the BFGS approximation.'
            )
            return self._calculate_function_and_derivatives_bfgs(iterate)

        return function_data

    def message(self) -> str:
        """Calculates and reports the proportion of calculation of the analytical
            hessian among all matrices involved

        :return: message reporting the proportion.
        :rtype: str
        """
        if self.number_of_matrices == 0:
            return ''

        actual_prop = (
            100
            * float(self.number_of_analytical_hessians)
            / float(self.number_of_matrices)
            if self.number_of_matrices != 0
            else 0
        )

        numerical = (
            f' [{self.number_of_numerical_issues} numerical issues]'
            if self.number_of_numerical_issues != 0
            else ''
        )

        the_message = (
            f'{self.number_of_analytical_hessians}/{self.number_of_matrices} '
            f'= {actual_prop:.1f}%{numerical}'
        )
        return the_message



"""File test_neighborhood.py

:author: Michel Bierlaire
:date: Wed Jul  5 15:58:49 2023

Tests for the neighborhood module
"""

import unittest
from unittest.mock import patch

import numpy as np

from biogeme_optimization.exceptions import (
    OptimizationError,
)  # Import the required exception class
from biogeme_optimization.neighborhood import OperatorsManagement


class TestOperatorsManagement(unittest.TestCase):
    def setUp(self):
        self.operators = {
            'operator1': lambda solution, neighborhood_size: solution
            + neighborhood_size,
            'operator2': lambda solution, neighborhood_size: solution
            - neighborhood_size,
        }
        self.op_management = OperatorsManagement(self.operators)

    def test_increase_score_last_operator(self):
        self.op_management.last_operator_name = 'operator1'
        self.op_management.increase_score_last_operator()
        self.assertEqual(self.op_management.scores['operator1'], 1)

    def test_increase_score_last_operator_unknown_operator(self):
        self.op_management.last_operator_name = 'unknown_operator'
        with self.assertRaises(OptimizationError):
            self.op_management.increase_score_last_operator()

    def test_decrease_score_last_operator(self):
        self.op_management.last_operator_name = 'operator1'
        self.op_management.scores['operator1'] = 2
        self.op_management.decrease_score_last_operator()
        self.assertEqual(self.op_management.scores['operator1'], 1)

    def test_decrease_score_last_operator_minimum_score(self):
        self.op_management.last_operator_name = 'operator1'
        self.op_management.scores['operator1'] = 0
        self.op_management.decrease_score_last_operator()
        self.assertEqual(self.op_management.scores['operator1'], -1)

    def test_decrease_score_last_operator_unknown_operator(self):
        self.op_management.last_operator_name = 'unknown_operator'
        with self.assertRaises(OptimizationError):
            self.op_management.decrease_score_last_operator()

    def test_select_operator(self):
        # Mocking np.random.choice to always return 'operator1'
        chosen = ('operator1',)
        with patch('numpy.random.choice', return_value=chosen):
            operator = self.op_management.select_operator()
            self.assertEqual(operator, self.operators['operator1'])

    def test_select_operator_with_unavailable_operators(self):
        self.op_management.available['operator2'] = False
        # Mocking np.random.choice to always return 'operator1'
        chosen = ('operator1',)
        with patch('numpy.random.choice', return_value=chosen):
            operator = self.op_management.select_operator()
            self.assertEqual(operator, self.operators['operator1'])

    def test_select_operator_with_min_probability(self):
        # Mocking np.random.choice to always return 'operator1'
        chosen = ('operator1',)
        with patch('numpy.random.choice', return_value=chosen):
            self.op_management.min_probability = 0.5
            operator = self.op_management.select_operator()
            self.assertEqual(operator, self.operators['operator1'])

    def test_select_operator_with_large_min_probability(self):
        with self.assertRaises(OptimizationError):
            self.op_management.min_probability = 0.7
            self.op_management.select_operator()

    def test_probability(self):
        self.op_management.available = [True, True, True]
        prob = np.array([0.0, 0.1, 0.45, 0.45])
        processed_prob = OperatorsManagement.enforce_minimum_probability(
            prob=prob, min_probability=0.2
        )
        expected_output = np.array([0, 0.2, 0.4, 0.4])
        np.testing.assert_array_equal(processed_prob, expected_output)

        prob = np.array([0.0, 0.5, 0.5])
        with self.assertRaises(OptimizationError):
            processed_prob = OperatorsManagement.enforce_minimum_probability(
                prob=prob, min_probability=0.6
            )

        prob = np.array([0.0, 0.5, 0.7])
        with self.assertRaises(OptimizationError):
            processed_prob = OperatorsManagement.enforce_minimum_probability(
                prob=prob, min_probability=0.6
            )

        prob = np.array([0.0, -0.5, 0.7])
        with self.assertRaises(OptimizationError):
            processed_prob = OperatorsManagement.enforce_minimum_probability(
                prob=prob, min_probability=0.6
            )

        prob = np.array([0.0, 0.1, 0.1, 0.8])
        with self.assertRaises(OptimizationError):
            processed_prob = OperatorsManagement.enforce_minimum_probability(
                prob=prob, min_probability=0.6
            )


if __name__ == '__main__':
    unittest.main()



"""File test_pareto.py

:author: Michel Bierlaire
:date: Tue Jul  4 17:25:58 2023

Tests for the pareto module
"""

import os
import tempfile
import unittest

from biogeme_optimization.exceptions import (
    OptimizationError,
)  # Import the required exception class
from biogeme_optimization.pareto import SetElement, Pareto


class TestSetElement(unittest.TestCase):
    def test_initialization(self):
        element_id = "element1"
        objectives = [1.0, 2.0, 3.0]
        set_element = SetElement(element_id, objectives)
        self.assertEqual(set_element.element_id, element_id)
        self.assertEqual(set_element.objectives, objectives)

    def test_initialization_missing_objectives(self):
        element_id = "element2"
        objectives = [1.0, None, 3.0]
        with self.assertRaises(OptimizationError):
            SetElement(element_id, objectives)

    def test_equality(self):
        element_id = "element3"
        objectives = [1.0, 2.0, 3.0]
        set_element1 = SetElement(element_id, objectives)
        set_element2 = SetElement(element_id, objectives)
        self.assertEqual(set_element1, set_element2)

    def test_equality_different_objectives(self):
        element_id = "element4"
        objectives1 = [1.0, 2.0, 3.0]
        objectives2 = [1.0, 3.0, 2.0]
        set_element1 = SetElement(element_id, objectives1)
        set_element2 = SetElement(element_id, objectives2)
        with self.assertRaises(OptimizationError):
            self.assertEqual(set_element1, set_element2)

    def test_hash(self):
        element_id = "element5"
        objectives = [1.0, 2.0, 3.0]
        set_element = SetElement(element_id, objectives)
        self.assertEqual(hash(set_element), hash((element_id, tuple(objectives))))

    def test_dominance(self):
        element_id1 = "element6"
        objectives1 = [1.0, 2.0, 3.0]
        element_id2 = "element7"
        objectives2 = [2.0, 3.0, 4.0]
        set_element1 = SetElement(element_id1, objectives1)
        set_element2 = SetElement(element_id2, objectives2)
        self.assertTrue(set_element1.dominates(set_element2))
        self.assertFalse(set_element2.dominates(set_element1))
        self.assertFalse(set_element1.dominates(set_element1))
        self.assertFalse(set_element2.dominates(set_element2))

    def test_dominance_incompatible_sizes(self):
        element_id1 = "element8"
        objectives1 = [1.0, 2.0, 3.0]
        element_id2 = "element9"
        objectives2 = [2.0, 3.0]
        set_element1 = SetElement(element_id1, objectives1)
        set_element2 = SetElement(element_id2, objectives2)
        with self.assertRaises(OptimizationError):
            set_element1.dominates(set_element2)


class TestPareto(unittest.TestCase):
    def setUp(self):
        self.filename = tempfile.mktemp()

    def tearDown(self):
        if os.path.exists(self.filename):
            os.remove(self.filename)

    def test_dump_restore(self):
        pareto = Pareto(self.filename)

        # Add some elements to the Pareto set
        pareto.add_invalid(SetElement('1', [1.0, 2.0]))
        pareto.add(SetElement('3', [5.0, 6.0]))
        pareto.add(SetElement('2', [3.0, 4.0]))

        # Check if the restored Pareto set matches the original one
        self.assertEqual(len(pareto.invalid), 1)
        self.assertEqual(len(pareto.pareto), 1)
        self.assertEqual(len(pareto.considered), 2)
        self.assertEqual(len(pareto.removed), 1)

        # Dump the Pareto set to a file
        pareto.dump()

        # Create a new Pareto object and restore from the file
        restored_pareto = Pareto(self.filename)
        restored_pareto.restore()

        # Check if the restored Pareto set matches the original one
        self.assertEqual(len(restored_pareto.invalid), 1)
        self.assertEqual(len(restored_pareto.pareto), 1)
        self.assertEqual(len(restored_pareto.considered), 2)
        self.assertEqual(len(restored_pareto.removed), 1)

        # Check the values of the restored elements
        self.assertSetEqual(pareto.pareto, restored_pareto.pareto)
        self.assertSetEqual(pareto.considered, restored_pareto.considered)
        self.assertSetEqual(pareto.removed, restored_pareto.removed)
        self.assertSetEqual(pareto.invalid, restored_pareto.invalid)

    def test_add(self):
        pareto = Pareto()

        # Add elements to the Pareto set
        pareto.add(SetElement('1', [1.0, 3.0]))
        pareto.add(SetElement('2', [2.0, 2.0]))
        pareto.add(SetElement('3', [3.0, 1.0]))

        # Check the size of the Pareto set
        self.assertEqual(pareto.length(), 3)
        self.assertEqual(pareto.length_of_all_sets(), (3, 3, 0, 0))

        # Add an element that is dominated by an existing element
        pareto.add(SetElement('4', [2.0, 4.0]))

        # Check that the dominated element was not added
        self.assertEqual(pareto.length(), 3)
        self.assertEqual(pareto.length_of_all_sets(), (3, 4, 0, 0))

        # Add an element that dominates an existing element
        pareto.add(SetElement('5', [2.5, 0.0]))

        # Check that the dominated element was removed
        self.assertEqual(pareto.length(), 3)
        self.assertEqual(pareto.length_of_all_sets(), (3, 5, 1, 0))

    def test_get_element_from_id(self):
        pareto = Pareto()

        # Add elements to the Pareto set
        pareto.add(SetElement('1', [1.0, 2.0]))
        pareto.add(SetElement('2', [3.0, 4.0]))
        pareto.add(SetElement('3', [5.0, 6.0]))

        # Get an element by ID
        element = pareto.get_element_from_id('2')

        # Check the properties of the element
        self.assertEqual(element.element_id, '2')
        self.assertEqual(element.objectives, [3.0, 4.0])

    def test_add_invalid(self):
        pareto = Pareto()

        # Add an invalid element to the Pareto set
        pareto.add_invalid(SetElement('1', [1.0, 2.0]))

        # Check the size of the Pareto set
        self.assertEqual(pareto.length(), 0)
        self.assertEqual(pareto.length_of_all_sets(), (0, 0, 0, 1))

        # Try adding the same invalid element again
        pareto.add_invalid(SetElement('1', [1.0, 2.0]))

        # Check that the element was not added again
        self.assertEqual(pareto.length(), 0)
        self.assertEqual(pareto.length_of_all_sets(), (0, 0, 0, 1))


if __name__ == '__main__':
    unittest.main()



"""File test_bfgs.py

:author: Michel Bierlaire
:date: Sat Jun 24 17:44:39 2023

Tests for the bfgs module
"""
import unittest
import numpy as np
import scipy.linalg as la
from biogeme_optimization.exceptions import (
    OptimizationError,
)  # Import the required exception class
from biogeme_optimization.bfgs import (
    bfgs,
    inverse_bfgs,
)
from biogeme_optimization.function import FunctionToMinimize


class BfgsTest(unittest.TestCase):
    def test_bfgs_1(self):
        # Test case 1: Basic example
        hessian_approx = np.array([[2, 0], [0, 2]])
        delta_x = np.array([1, 1])
        delta_g = np.array([1, 1])

        expected_result = np.array([[1.5, -0.5], [-0.5, 1.5]])

        result = bfgs(hessian_approx, delta_x, delta_g)
        np.testing.assert_array_almost_equal(result, expected_result, decimal=6)

    def test_bfgs_2(self):
        # Test case 2: Larger matrices
        hessian_approx = np.array([[2, 1, 0], [1, 3, -1], [0, -1, 4]])
        delta_x = np.array([1, -1, 1])
        delta_g = np.array([4, 0, -1])

        expected_result = np.array(
            [
                [7.22222222, 1.33333333, -1.88888888],
                [1.33333333, 2.0, 0.66666666],
                [-1.88888888, 0.66666666, 1.55555555],
            ]
        )

        result = bfgs(hessian_approx, delta_x, delta_g)
        np.testing.assert_array_almost_equal(result, expected_result, decimal=6)

    def test_bfgs_3(self):
        # Test case 3:delta_x and delta_g are orthogonal
        hessian_approx = np.array([[2, 0], [0, 2]])
        delta_x = np.array([1, -1])
        delta_g = np.array([-1, 1])

        expected_result = np.array([[1.2, 0.8], [0.8, 1.2]])

        result = bfgs(hessian_approx, delta_x, delta_g)
        np.testing.assert_array_almost_equal(result, expected_result, decimal=6)

    def test_inverse_bfgs_1(self):
        # Test case 1: Basic example
        hessian_approx = np.array([[2, 0], [0, 2]])
        inverse_hessian_approx = np.linalg.inv(hessian_approx)
        delta_x = np.array([1, 1])
        delta_g = np.array([1, 1])

        expected_result_bfgs = np.array([[1.5, -0.5], [-0.5, 1.5]])
        expected_result = np.linalg.inv(expected_result_bfgs)

        result = inverse_bfgs(inverse_hessian_approx, delta_x, delta_g)
        np.testing.assert_array_almost_equal(result, expected_result, decimal=6)

    def test_inverse_bfgs_2(self):
        # Test case 2: Larger matrices
        hessian_approx = np.array([[2, 1, 0], [1, 3, -1], [0, -1, 4]])
        inverse_hessian_approx = np.linalg.inv(hessian_approx)
        delta_x = np.array([1, -1, 1])
        delta_g = np.array([4, 0, -1])

        expected_result_bfgs = np.array(
            [
                [7.22222222, 1.33333333, -1.88888888],
                [1.33333333, 2.0, 0.66666666],
                [-1.88888888, 0.66666666, 1.55555555],
            ]
        )
        expected_result = np.linalg.inv(expected_result_bfgs)
        result = inverse_bfgs(inverse_hessian_approx, delta_x, delta_g)
        np.testing.assert_array_almost_equal(result, expected_result, decimal=6)

    def test_inverse_bfgs_3(self):
        # Test case 3:delta_x and delta_g are orthogonal. Note that,
        # in that case, the "inverse" formula is not valid.
        hessian_approx = np.array([[2, 0], [0, 2]])
        inverse_hessian_approx = np.linalg.inv(hessian_approx)
        delta_x = np.array([1, -1])
        delta_g = np.array([-1, 1])

        with self.assertRaises(OptimizationError):
            result = inverse_bfgs(inverse_hessian_approx, delta_x, delta_g)


if __name__ == '__main__':
    unittest.main()



"""File bounds.py.

:author: Michel Bierlaire
:date: Wed Jun 21 09:56:40 2023

Class in charge of the management of bound constraints
"""

from __future__ import annotations
import logging
import numpy as np

from biogeme_optimization.exceptions import OptimizationError
from biogeme_optimization.diagnostics import ConjugateGradientDiagnostic

logger: logging.Logger = logging.getLogger(__name__)


def inner_product_subspace(
    vector_a: np.ndarray, vector_b: np.ndarray, subspace_indices
) -> float:
    """Given two input vectors `vector_a` and `vector_b`, this
    function calculates the inner product of the entries in `vector_a`
    and `vector_b` that correspond to the provided `subspace_indices`.

    :param vector_a: The first input vector.
    :type vector_a: numpy.ndarray

    :param vector_b: The second input vector.
    :type vector_b: numpy.ndarray

    :param subspace_indices: The indices of the selected entries.
    :type subspace_indices: iterable

    :return: The inner product of the selected entries.
    :rtype: float

    :raises IndexError: If the `subspace_indices` are out of bounds
        for `vector_a` or `vector_b`.

    """

    # Extract the selected entries from the input vectors
    selected_a = vector_a[list(subspace_indices)]
    selected_b = vector_b[list(subspace_indices)]

    # Calculate the inner product of the selected entries
    inner_product = np.dot(selected_a, selected_b)

    return inner_product


def matrix_vector_mult_subspace(
    matrix: np.ndarray, vector: np.ndarray, subspace_indices
) -> np.ndarray:
    """Performs matrix-vector multiplication involving selected entries of the vector.

    Given a square matrix and a vector, this function calculates the multiplication
    between the selected columns of the matrix and the selected entries of the vector
    defined by a set of indices.

    :param matrix: Input square matrix of shape (n, n).
    :type matrix: numpy.ndarray

    :param vector: Input vector of length n.
    :type vector: numpy.ndarray

    :param subspace_indices: Set of indices specifying the selected
        entries of the vector and columns of the matrix.

    :type subspace_indices: set or numpy.ndarray

    :return: Resulting vector of length n.
    :rtype: numpy.ndarray

    :raises ValueError: If the dimensions of the matrix and vector are incompatible.
    :raises IndexError: If any index in subspace_indices is out of range.

    """
    if matrix.shape[1] != vector.shape[0]:
        raise ValueError(
            f'Incompatible dimensions between matrix [{matrix.shape[1]}] '
            f'and vector [{vector.shape[0]}]'
        )

    if any(index >= vector.shape[0] for index in subspace_indices):
        raise IndexError(
            'Subspace index out of range for the vector of size {vector.shape[0]}'
        )

    # Extract the selected entries from the input vector
    selected_vector = vector[list(subspace_indices)] if subspace_indices else vector

    # Select the columns of the matrix based on the subspace indices
    selected_matrix = matrix[:, list(subspace_indices)] if subspace_indices else matrix

    # Perform matrix-vector multiplication using the selected entries and columns
    result = np.dot(selected_matrix, selected_vector)

    return result


class Bounds:
    """This class is designed for the management of simple bound constraints."""

    INFINITY = np.sqrt(np.finfo(float).max)

    def __init__(self, bounds: list[tuple[float | None, float | None]]):
        """Constructor.

        :param bounds: list of tuples (ell,u) containing the lower and
                  upper bounds for each free parameter.

        :type bounds: list(tuple)

        :raises biogeme.exceptions.OptimizationError: if the bounds are incompatible

        """

        def none_to_minus_infinity(value: float) -> float:
            """Returns -infinity if the input is None, otherwise returns the input canonical_value

            :param value: The input canonical_value.
            :type value: float

            :return: The modified canonical_value. If x is None, returns
                -infinity; otherwise, returns x.
            :rtype: float

            """
            if value is None:
                return -self.INFINITY

            return value

        def none_to_plus_infinity(value: float) -> float:
            """Returns +infinity if the input is None, otherwise
                returns the input canonical_value.

            :param value: The input canonical_value.
            :type value: float

            :return: The modified canonical_value. If x is None, returns
                +infinity; otherwise, returns x.
            :rtype: float

            """
            if value is None:
                return self.INFINITY

            return value

        self.bounds = bounds
        """list of tuples (ell, u) containing the lower and upper bounds for
            each free parameter.
        """

        self.dimension = len(
            bounds
        )  #: number of optimization unsorted_set_of_variables

        self.lower_bounds = np.array(
            [none_to_minus_infinity(bound[0]) for bound in bounds]
        )
        """ List of lower bounds """

        self.upper_bounds = np.array(
            [none_to_plus_infinity(bound[1]) for bound in bounds]
        )
        """ List of upper bounds """

        wrong_bounds = np.array(
            [
                bound[0] > bound[1]
                for bound in bounds
                if bound[0] is not None and bound[1] is not None
            ]
        )
        if wrong_bounds.any():
            error_msg = (
                f'Incompatible bounds for indice(s) '
                f'{np.nonzero(wrong_bounds)[0]}: '
                f'{[bounds[i] for i in np.nonzero(wrong_bounds)[0]]}'
            )
            raise OptimizationError(error_msg)

    @classmethod
    def from_bounds(cls, lower: np.ndarray, upper: np.ndarray) -> Bounds:
        """Ctor using two vectors of bounds

        :param lower: vector of lower bounds
        :type lower: numpy.array

        :param upper: vector of upper bounds
        :type upper: numpy.array

        """
        if len(lower) != len(upper):
            error_msg = f'Inconsistent length of arrays: {len(lower)} and {len(upper)}'
            raise OptimizationError(error_msg)

        list_of_tuples = list(zip(lower, upper))
        return cls(list_of_tuples)

    def __str__(self) -> str:
        """Magic string."""
        return self.bounds.__str__()

    def __repr__(self) -> str:
        return self.bounds.__repr__()

    def project(self, point: np.ndarray) -> np.ndarray:
        """Project a point onto the feasible domain defined by the bounds.

        :param point: point to project
        :type point: numpy.array

        :return: projected point
        :rtype: numpy.array

        :raises biogeme.exceptions.OptimizationError: if the dimensions
                are inconsistent
        """

        if len(point) != self.dimension:
            raise OptimizationError(
                f'Incompatible size: {len(point)}' f' and {self.dimension}'
            )

        new_point = np.clip(point, self.lower_bounds, self.upper_bounds)
        return new_point

    def intersect(self, other_bounds: Bounds) -> Bounds:
        """
        Create a bounds object representing the intersection of two regions.

        :param other_bounds: other bound object that must be intersected.
        :type other_bounds: class Bounds

        :return:  bound object, intersection of the two.
        :rtype: class Bounds

        :raises biogeme.exceptions.OptimizationError: if the dimensions
              are inconsistent
        """

        if other_bounds.dimension != self.dimension:
            raise OptimizationError(
                f'Incompatible size: {other_bounds.dimension} ' f'and {self.dimension}'
            )

        new_bounds = [
            (
                np.maximum(self.lower_bounds[i], other_bounds.lower_bounds[i]),
                np.minimum(self.upper_bounds[i], other_bounds.upper_bounds[i]),
            )
            for i in range(self.dimension)
        ]

        # Note that an exception with be raised at the creation of the
        # new 'Bounds' object is the bounds are incompatible.

        result = Bounds(new_bounds)
        return result

    def intersection_with_trust_region(
        self, point: np.ndarray, radius: float
    ) -> Bounds:
        """Create a Bounds object representing the intersection
            between the bounds and the trust region:
             point-radius <= x <= point+radius
            and
             ell <= x <= u

        :param point: center of the trust region
        :type point: numpy.array

        :param radius: radius of the trust region (infinity norm)
        :type radius: float

        :return: intersection between the feasible region and the trust region
        :rtype: class Bounds

        :raises OptimizationError: if the point is not feasible

        """
        if not self.feasible(point):
            error_msg = f'Center of the trust region of radius {radius} is not feasible'
            raise OptimizationError(error_msg)
        trust_region = Bounds.from_bounds(point - radius, point + radius)
        return self.intersect(trust_region)

    def get_bounds_for_trust_region_subproblem(
        self, point: np.ndarray, radius: float
    ) -> Bounds:
        """Create a Bounds object representing the bounds on the step
            for the trust region sub-problem. It is the intersection between:
              -radius <= d <= radius
            and
              ell-point <= d <= u-point

        :param point: center of the trust region
        :type point: numpy.array

        :param radius: radius of the trust region (infinity norm)
        :type radius: float

        :return: intersection between the feasible region and the trust region
        :rtype: class Bounds

        :raises OptimizationError: if the dimensions are inconsistent

        """
        if len(point) != self.dimension:
            raise OptimizationError(
                f'Incompatible size: {len(point)}' f' and {self.dimension}'
            )
        trust_region = Bounds.from_bounds(
            np.full(self.dimension, -radius),
            np.full(self.dimension, radius),
        )
        shifted_bounds = Bounds.from_bounds(
            self.lower_bounds - point, self.upper_bounds - point
        )
        return shifted_bounds.intersect(trust_region)

    def subspace(self, selected_variables: np.ndarray) -> Bounds:
        """Generate a Bounds object for selected unsorted_set_of_variables

        :param selected_variables: boolean vector. If an entry is True,
                   the corresponding unsorted_set_of_variables is considered.
        :type selected_variables: numpy.array(bool)

        :return: bound object
        :rtype: class Bounds

        :raises biogeme.exceptions.OptimizationError: if the dimensions
                are inconsistent
        """
        if len(selected_variables) != self.dimension:
            raise OptimizationError(
                f'Incompatible size: ' f'{len(selected_variables)} and {self.dimension}'
            )

        selected_bounds = [
            bound
            for bound, selected in zip(self.bounds, selected_variables)
            if selected
        ]
        return Bounds(selected_bounds)

    def feasible(self, point: np.ndarray) -> bool:
        """Check if point verifies the bound constraints

        :param point: point to project
        :type point: numpy.array

        :return: True if point is feasible, False otherwise.
        :rtype: bool

        :raises OptimizationError: if the dimensions are inconsistent
        """
        if point.size != self.dimension:
            raise OptimizationError(
                f'Incompatible size: ' f'point:{point.size} and {self.dimension}'
            )

        if np.any(self.lower_bounds - point > np.finfo(float).eps):
            return False
        if np.any(point - self.upper_bounds > np.finfo(float).eps):
            return False
        return True

    def maximum_step(
        self, point: np.ndarray, direction: np.ndarray
    ) -> tuple[float, np.ndarray]:
        """Calculates the maximum step that can be performed
        along a direction while staying feasible.

        :param point: reference point
        :type point: numpy.array

        :param direction: direction
        :type direction: numpy.array

        :return: the largest alpha such that point + alpha * direction is feasible
                 and the list of indices achieving this canonical_value.
        :rtype: float, int

        :raises biogeme.exceptions.OptimizationError: if the point is infeasible

        """

        if not self.feasible(point):
            raise OptimizationError(f'Infeasible point: {point}')

        def calculate_alpha(
            point_element: np.ndarray,
            lower: np.ndarray,
            upper: np.ndarray,
            direction_element: np.ndarray,
        ) -> np.ndarray:
            """Calculate the canonical_value of alpha on numpy arrays"""
            result = np.empty_like(point_element, dtype=float)

            upper_bound_exists = upper != np.inf
            lower_bound_exists = lower != -np.inf
            strictly_positive_entries = direction_element > np.finfo(float).eps
            strictly_negative_entries = direction_element < -np.finfo(float).eps

            # If the upper bound exist and the direction is positive,
            # we calculate the distance to the bound
            mask1 = upper_bound_exists & strictly_positive_entries
            result[mask1] = (upper[mask1] - point_element[mask1]) / direction_element[
                mask1
            ]

            # If there is no upper bound, and the direction is
            # positive, the step is infinite.
            mask2 = ~upper_bound_exists & strictly_positive_entries
            result[mask2] = np.inf

            # If the lower bound exist, and the direction is negative,
            # we calculate the distance to the bound.
            mask3 = lower_bound_exists & strictly_negative_entries
            result[mask3] = (lower[mask3] - point_element[mask3]) / direction_element[
                mask3
            ]

            # If the lower bound does not exist, and the direction is
            # negative, the step is infinite.
            mask4 = ~lower_bound_exists & strictly_negative_entries
            result[mask4] = np.inf

            # If the direction is zero, or very close to zero, the
            # step is infinite
            mask5 = ~strictly_positive_entries & ~strictly_negative_entries
            result[mask5] = np.inf

            return result

        alpha = calculate_alpha(point, self.lower_bounds, self.upper_bounds, direction)
        the_minimum = np.amin(alpha)
        return the_minimum, np.where(alpha == the_minimum)[0]

    def activity(
        self, point: np.ndarray, epsilon: float = np.finfo(float).eps
    ) -> np.ndarray:
        """Determines the activity status of each variable.

        :param point: point for which the activity must be determined.
        :type point: numpy.array

        :param epsilon: a bound is considered active if the distance
                        to it is less than epsilon.
        :type epsilon: float

        :return: a vector, same length as point, where each entry reports
            the activity of the corresponding variable:

            - 0 if no bound is active
            - -1 if the lower bound is active
            - 1 if the upper bound is active

        :rtype: numpy.array

        :raises biogeme.exceptions.OptimizationError: if the vector point is
                    not feasible

        :raises biogeme.exceptions.OptimizationError: if the dimensions
                    of x and bounds do not match.

        """
        if len(point) != self.dimension:
            raise OptimizationError(
                f'Incompatible size: {len(point)}' f' and {self.dimension}'
            )
        if not self.feasible(point):
            raise OptimizationError(
                f'{point} is not feasible for the ' f'bounds {self.bounds}'
            )

        activity_status = np.zeros(self.dimension, dtype=int)
        lb_mask = np.logical_and(
            self.lower_bounds != -np.inf, point - self.lower_bounds <= epsilon
        )
        ub_mask = np.logical_and(
            self.upper_bounds != np.inf, self.upper_bounds - point <= epsilon
        )

        activity_status[lb_mask] = -1
        activity_status[ub_mask] = 1

        return activity_status

    def active_constraints(
        self, point: np.ndarray, epsilon: float = np.finfo(float).eps
    ) -> set[int]:
        """Determines the indices of active constraints for a given point.

        :param point: The point for which to determine the active constraints.
        :type point: numpy.array

        :param epsilon: The tolerance used to determine the activity
            status of constraints.  the_matrix constraint is considered active
            if the distance to it is less than epsilon.
        :type epsilon: float, optional

        :return: the_matrix set containing the indices of active constraints.
        :rtype: set

        :raises biogeme.exceptions.OptimizationError: If the
            dimensions are inconsistent.
        :raises biogeme.exceptions.OptimizationError: If the point is
            infeasible.
        """

        activity_status = self.activity(point, epsilon)
        return {i for i, status in enumerate(activity_status) if status != 0}

    def inactive_constraints(
        self, point: np.ndarray, epsilon: float = np.finfo(float).eps
    ) -> set[int]:
        """Determines the indices of inactive constraints for a given point.

        :param point: The point for which to determine the active constraints.
        :type point: numpy.array

        :param epsilon: The tolerance used to determine the activity
            status of constraints.  the_matrix constraint is considered active
            if the distance to it is less than epsilon.
        :type epsilon: float, optional

        :return: the_matrix set containing the indices of inactive constraints.
        :rtype: set

        :raises biogeme.exceptions.OptimizationError: If the
            dimensions are inconsistent.
        :raises biogeme.exceptions.OptimizationError: If the point is
            infeasible.

        """

        activity_status = self.activity(point, epsilon)
        return {i for i, status in enumerate(activity_status) if status == 0}

    def breakpoints(
        self, point: np.ndarray, direction: np.ndarray
    ) -> list[tuple[int, float]]:
        """Projects the direction direction, starting from point,
        on the intersection of the bound constraints

        :param point: current point
        :type point: numpy.array

        :param direction: search direction
        :type direction: numpy.array

        :return: list of tuple (index, canonical_value), where index is the index
                 of the variable, and canonical_value the canonical_value of the corresponding
                 breakpoint.
        :rtype: list(tuple(int,float))

        :raises biogeme.exceptions.OptimizationError: if the dimensions
                are inconsistent

        :raises biogeme.exceptions.OptimizationError: if point is infeasible

        """
        if len(direction) != self.dimension:
            raise OptimizationError(
                f'Incompatible size: {self.dimension}' f' and {len(direction)}'
            )

        if not self.feasible(point):
            raise OptimizationError('Infeasible point')

        positive_mask = direction > np.finfo(float).eps
        negative_mask = direction < -np.finfo(float).eps

        upper_valid = self.upper_bounds != np.inf
        lower_valid = self.lower_bounds != -np.inf

        upper_breaks = np.where(
            positive_mask & upper_valid, (self.upper_bounds - point) / direction, np.inf
        )
        lower_breaks = np.where(
            negative_mask & lower_valid, (self.lower_bounds - point) / direction, np.inf
        )

        break_points = [(i, val) for i, val in enumerate(upper_breaks) if val < np.inf]
        break_points.extend(
            [(i, val) for i, val in enumerate(lower_breaks) if val < np.inf]
        )
        if not break_points:
            raise OptimizationError("Infeasible point")

        return sorted(break_points, key=lambda the_point: the_point[1])

    def projected_direction(
        self, x_current: np.ndarray, direction: np.ndarray
    ) -> np.ndarray:
        """Calculates the projected direction.

        Given the current point x_current and the direction,
        the function determines the projected direction:

        .. math:: \\lim_{\\alpha \to 0^+} P[x + \\alpha d] - x

        :param x_current: current point
        :type x_current: numpy.array

        :param direction: direction to follow
        :type direction: numpy.array

        :return: the projected direction
        :rtype: numpy.array

        :raises OptimizationError: if the dimensions are inconsistent
        :raises OptimizationError: if the current point is infeasible

        """
        activity_status = self.activity(x_current)
        beyond_upper = np.logical_and(activity_status == 1, direction > 0)
        beyond_lower = np.logical_and(activity_status == -1, direction < 0)
        projected_direction = direction
        projected_direction[beyond_upper] = 0
        projected_direction[beyond_lower] = 0

        return projected_direction

    def generalized_cauchy_point(
        self, x_current: np.ndarray, g_current: np.ndarray, h_current: np.ndarray
    ) -> np.ndarray:
        """Implementation of Step 2 of the Specific Algorithm by
        `Conn et al. (1988)`_.

        .. _`Conn et al. (1988)`: https://www.ams.org/journals/mcom/1988-50-182/S0025-5718-1988-0929544-3/S0025-5718-1988-0929544-3.pdf

        The quadratic model is defined as

        .. math:: m(x) = f(x_k) + (x - x_k)^T g_k +
                  \\frac{1}{2} (x-x_k)^T H (x-x_k).

        :param x_current: current point
        :type x_current: numpy.array. Dimension n.

        :param g_current: vector g involved in the quadratic model definition.
        :type g_current: numpy.array. Dimension n.

        :param h_current: matrix h_current involved in the quadratic model definition.
        :type h_current: numpy.array. Dimension n x n.

        :return: generalized Cauchy point based on inexact line search.
        :rtype: numpy.array. Dimension n.

        :raises biogeme.exceptions.OptimizationError: if the dimensions are
               inconsistent
        :raises biogeme.exceptions.OptimizationError: if xk is infeasible
        """
        if not isinstance(x_current, np.ndarray):
            raise OptimizationError(f'Not an array {x_current}: {type(x_current)}')

        if len(x_current) != self.dimension:
            raise OptimizationError(
                f'Incompatible size: {len(x_current)}' f' and {self.dimension}'
            )
        if len(g_current) != self.dimension:
            raise OptimizationError(
                f'Incompatible size: {len(g_current)}' f' and {self.dimension}'
            )
        if np.isnan(g_current).any():
            raise OptimizationError(f'Invalid gradient: {g_current}')

        if h_current.shape[0] != self.dimension or h_current.shape[1] != self.dimension:
            raise OptimizationError(
                f'Incompatible size: {h_current.shape}' f' and {self.dimension}'
            )

        if np.isnan(h_current).any():
            raise OptimizationError(f'Invalid hessian: {h_current}')

        if not self.feasible(x_current):
            raise OptimizationError(
                f'Infeasible iterate {x_current} '
                '[LB: {self.lower_bounds} UB: {self.upper_bounds}]'
            )

        x = x_current
        g = g_current - h_current @ x_current
        direction = self.projected_direction(x_current, -g_current)

        fprime = np.inner(g_current, direction)
        if fprime >= 0:
            if self.dimension <= 10:
                logger.warning(
                    f'GCP: direction {direction} is not a descent direction at {x}.'
                )
            else:
                logger.warning('GCP: not a descent direction.')

        f_second = np.inner(direction, h_current @ direction)

        set_active_constraints = set()

        while len(set_active_constraints) < self.dimension:
            # Calculate the maximum step that can be done in the current direction
            delta_t, _ = self.maximum_step(x, direction)
            # Test whether the GCP has been found
            if f_second > 0:
                ratio = -fprime / f_second
                if 0 < ratio < delta_t:
                    x = x + ratio * direction
                    return x

            # In theory, x + delta_t * direction must be feasible. However,
            # there may be some numerical problem. Therefore, we
            # project it on the feasible domain to make sure to obtain
            # a feasible point.
            x_plus = self.project(x + delta_t * direction)

            # Note that there may be more than one variable activated
            # at x_plus. Therefore, we recalculate the activity
            # status.
            set_active_constraints_plus = self.active_constraints(x_plus)
            activated = set_active_constraints_plus - set_active_constraints

            x = x_plus
            set_active_constraints = set_active_constraints_plus

            # Update line derivatives
            bd = np.zeros_like(direction)
            bd[list(activated)] = direction[list(activated)]
            b = h_current @ bd

            d_dot_g = np.sum([direction[i] * g[i] for i in activated])
            fprime += delta_t * f_second - np.inner(b, x) - d_dot_g
            f_second += np.inner(b, bd - 2 * direction)
            direction[list(activated)] = 0.0
            if np.all(direction == 0) or fprime >= 0:
                return x

        return x

    def truncated_conjugate_gradient(
        self, gradient: np.ndarray, hessian: np.ndarray, tol: float = 1.0e-6
    ) -> tuple[np.ndarray, ConjugateGradientDiagnostic]:
        """Find an approximation of the trust region sub-problem using
            the truncated conjugate gradient method, where the step d
            mus verify the bound constraints.

        :param gradient: gradient of the quadratic model.
        :type gradient: numpy.array

        :param hessian: hessian of the quadratic model.
        :type hessian: numpy.array

        :param tol: tolerance on the norm of the gradient, used as a stopping criterion
        :type tol: float

        :return: d, diagnostic, where

              - d is the approximate solution of the trust region sub-problem,
              - diagnostic is the nature of the solution:

                * CONVERGENCE for convergence,
                * OUT_OF_TRUST_REGION if out of the trust region,
                * NEGATIVE_CURVATURE if negative curvature detected.
                * NUMERICAL_PROBLEM if a numerical problem has been encountered

        :rtype: numpy.array, ConjugateGradientDiagnostic

        """
        dimension = gradient.size
        if hessian.shape[0] != dimension:
            error_msg = f'Incompatible sizes: hessian:{hessian.shape[0]} and gradient:{dimension}'
            raise OptimizationError(error_msg)

        if hessian.shape[0] != hessian.shape[1]:
            error_msg = (
                f'Matrix must be square and not {hessian.shape[0]} x {hessian.shape[1]}'
            )
            raise OptimizationError(error_msg)

        xk = np.zeros(dimension)
        hx_plus_g = gradient
        dk = -gradient
        for _ in range(dimension):
            try:
                if np.linalg.norm(hx_plus_g) <= tol:
                    diagnostic = ConjugateGradientDiagnostic.CONVERGENCE
                    step = xk
                    return step, diagnostic

                curvature = np.inner(dk, hessian @ dk)
                if curvature <= 0:
                    # Negative curvature has been detected
                    diagnostic = ConjugateGradientDiagnostic.NEGATIVE_CURVATURE
                    step, _ = self.maximum_step(xk, dk)
                    solution = xk + step * dk
                    return solution, diagnostic
                alpha_k = -np.inner(dk, hx_plus_g) / curvature
                xkp1 = xk + alpha_k * dk
                if np.isnan(xkp1).any() or not self.feasible(xkp1):
                    # Out of the trust region
                    diagnostic = ConjugateGradientDiagnostic.OUT_OF_TRUST_REGION
                    step, _ = self.maximum_step(xk, dk)
                    solution = xk + step * dk
                    return solution, diagnostic
                xk = xkp1
                next_hx_plus_g = hessian @ xk + gradient
                beta_k = np.inner(next_hx_plus_g, next_hx_plus_g) / np.inner(
                    hx_plus_g, hx_plus_g
                )
                dk = -next_hx_plus_g + beta_k * dk
                hx_plus_g = next_hx_plus_g
            except ValueError:
                # Numerical problem. We follow the last direction
                # until the border of the trust region
                diagnostic = ConjugateGradientDiagnostic.NUMERICAL_PROBLEM
                step, _ = self.maximum_step(xk, dk)
                solution = xk + step * dk
                return solution, diagnostic
        # In theory, after n iterations, the algorithm has converged
        diagnostic = ConjugateGradientDiagnostic.CONVERGENCE
        step = xk
        return step, diagnostic

    def truncated_conjugate_gradient_subspace(
        self,
        iterate: np.ndarray,
        gradient: np.ndarray,
        hessian: np.ndarray,
        radius: float,
        tol: float = np.finfo(np.float64).eps ** 0.3333,
    ) -> tuple[np.ndarray, ConjugateGradientDiagnostic]:
        """Find an approximation of the solution of the trust region
        subproblem using the truncated conjugate gradient method within
        the subspace of free unsorted_set_of_variables. Free unsorted_set_of_variables are those
        corresponding to inactive constraints at the generalized Cauchy
        point.

        :param iterate: current iterate.
        :type iterate: numpy.array

        :param gradient: gradient of the quadratic model.
        :type gradient: numpy.array

        :param hessian: hessian of the quadrartic model.
        :type hessian: numpy.array

        :param radius: radius of the trust region.
        :type radius: float

        :param tol: tolerance used for stopping criterion.
        :type tol: float

        :return: d, diagnostic, where

              - d is the approximate solution of the trust region subproblem,
              - diagnostic is the nature of the solution:

                * CONVERGENCE for convergence,
                * OUT_OF_TRUST_REGION if out of the trust region,
                * NEGATIVE_CURVATURE if negative curvature detected.
                * NUMERICAL_PROBLEM if a numerical problem has been encountered

        :rtype: numpy.array, ConjugateGradientDiagnostic

        :raises OptimizationError: if the dimensions are inconsistent

        """
        if (
            np.isnan(iterate).any()
            or np.isnan(gradient).any()
            or np.isnan(hessian).any()
        ):
            raise OptimizationError(
                'Invalid input: iterate, gradient, or hessian contains NaN values.'
            )

        # For numerical reasons, it is better to project the iterate
        # on the bounds, although it should be feasible.
        iterate = self.project(iterate)

        # First, we calculate the intersection between the trust region on
        # the bounds. The trust region is also a bound constraint (based
        # on infinity norm) of radius 'radius', centered at iterate.
        intersection = self.intersection_with_trust_region(iterate, radius)

        # Then, we calculate the generalized Cauchy point
        gcp = intersection.generalized_cauchy_point(iterate, gradient, hessian)

        bounds_trust_region_subproblem = self.get_bounds_for_trust_region_subproblem(
            iterate, radius
        )

        # We fix the unsorted_set_of_variables active at the GCP
        activity_status: np.ndarray = intersection.activity(gcp)
        free_variables: np.ndarray = activity_status == 0
        fixed_variables: np.ndarray = activity_status != 0

        if not free_variables.any():
            return gcp, ConjugateGradientDiagnostic.CONVERGENCE

        bounds_subspace = bounds_trust_region_subproblem.subspace(free_variables)
        gradient_subspace = gradient[free_variables]
        hessian_subspace = hessian[free_variables][:, free_variables]

        candidate, diagnostic = bounds_subspace.truncated_conjugate_gradient(
            gradient_subspace, hessian_subspace, tol
        )

        solution = gcp
        solution[free_variables] = iterate[free_variables] + candidate
        return solution, diagnostic



"""File test_bounds.py

:author: Michel Bierlaire
:date: Wed Jun 21 10:06:46 2023

Tests for the bounds module
"""

import unittest
import numpy as np
from biogeme_optimization.exceptions import (
    OptimizationError,
)  # Import the required exception class
from biogeme_optimization.bounds import (
    Bounds,
    inner_product_subspace,
    matrix_vector_mult_subspace,
)
from biogeme_optimization.diagnostics import ConjugateGradientDiagnostic


class TestCode(unittest.TestCase):
    def setUp(self):
        the_bounds = [(1, 10), (2, 20), (3, 30)]
        self.bounds = Bounds(the_bounds)

    def test_constructor(self):
        same_bounds = Bounds.from_bounds(
            self.bounds.lower_bounds, self.bounds.upper_bounds
        )
        np.testing.assert_array_equal(
            self.bounds.lower_bounds, same_bounds.lower_bounds
        )
        np.testing.assert_array_equal(
            self.bounds.upper_bounds, same_bounds.upper_bounds
        )

    def test_inner_product_subspace(self):
        a = np.array([1, 2, 3, 4, 5])
        b = np.array([2, 4, 6, 8, 10])
        subspace_indices = [1, 3, 4]

        result = inner_product_subspace(a, b, subspace_indices)
        expected = 2 * 4 + 4 * 8 + 5 * 10

        self.assertEqual(result, expected)

    def test_inner_product_subspace_empty_indices(self):
        a = np.array([1, 2, 3, 4, 5])
        b = np.array([2, 4, 6, 8, 10])
        subspace_indices = []

        result = inner_product_subspace(a, b, subspace_indices)
        expected = 0  # Empty selection, inner product is 0

        self.assertEqual(result, expected)

    def test_inner_product_subspace_invalid_indices(self):
        a = np.array([1, 2, 3, 4, 5])
        b = np.array([2, 4, 6, 8, 10])
        subspace_indices = [1, 5]  # Out of bounds index

        with self.assertRaises(IndexError):
            inner_product_subspace(a, b, subspace_indices)

    def test_matrix_vector_mult_subspace(self):
        # Test case 1
        matrix = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        vector = np.array([1, 2, 3])
        subspace_indices = {0, 2}
        expected_result = np.array([1 * 1 + 3 * 3, 4 * 1 + 6 * 3, 7 * 1 + 9 * 3])
        result = matrix_vector_mult_subspace(matrix, vector, subspace_indices)
        np.testing.assert_array_equal(result, expected_result)

        # Test case 2
        matrix = np.array([[2, 4, 6], [1, 3, 5], [0, 7, 9]])
        vector = np.array([-1, 2, 0])
        subspace_indices = {1, 2}
        expected_result = np.array([4 * 2 + 6 * 0, 3 * 2 + 5 * 0, 7 * 2 + 9 * 0])
        result = matrix_vector_mult_subspace(matrix, vector, subspace_indices)
        np.testing.assert_array_equal(result, expected_result)

        # Test case 3 (empty subspace_indices)
        matrix = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        vector = np.array([1, 2, 3])
        subspace_indices = set()
        expected_result = np.array(
            [
                1 * 1 + 2 * 2 + 3 * 3,
                4 * 1 + 5 * 2 + 6 * 3,
                7 * 1 + 8 * 2 + 9 * 3,
            ]
        )
        result = matrix_vector_mult_subspace(matrix, vector, subspace_indices)
        np.testing.assert_array_equal(result, expected_result)

    def test_invalid_dimensions(self):
        # Test case with incompatible dimensions
        matrix = np.array([[1, 2], [4, 5]])
        vector = np.array([1, 2, 3])
        subspace_indices = [0, 2]
        with self.assertRaises(ValueError):
            matrix_vector_mult_subspace(matrix, vector, subspace_indices)

    def test_invalid_indices(self):
        # Test case with out-of-range indices
        matrix = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        vector = np.array([1, 2, 3])
        subspace_indices = [0, 3]  # Index 3 is out of range
        with self.assertRaises(IndexError):
            matrix_vector_mult_subspace(matrix, vector, subspace_indices)

    def test_project_infinite(self):
        the_bounds = Bounds([(0, 1), (0, None), (None, 1), (None, None)])
        point_1 = np.array([-1, -1, -1, -1])
        output_1 = the_bounds.project(point_1)
        expected_1 = np.array([0, 0, -1, -1])
        np.testing.assert_array_equal(output_1, expected_1)

        point_2 = np.array([2, 2, 2, 2])
        output_2 = the_bounds.project(point_2)
        expected_2 = np.array([1, 2, 1, 2])
        np.testing.assert_array_equal(output_2, expected_2)

    def test_project(self):
        example_input = [5, 15, 25]
        expected_output = np.array(
            [
                5,
                15,
                25,
            ]
        )  # Expected output when all values are within bounds
        np.testing.assert_array_equal(
            self.bounds.project(example_input), expected_output
        )

    def test_lower_out_of_bounds(self):
        input_with_lower_out_of_bounds = [0, 15, 25]
        expected_output_with_lower_out_of_bounds = np.array([1, 15, 25])
        np.testing.assert_array_almost_equal(
            self.bounds.project(input_with_lower_out_of_bounds),
            expected_output_with_lower_out_of_bounds,
        )

    def test_upper_out_of_bounds(self):
        input_with_upper_out_of_bounds = [5, 25, 40]
        expected_output_with_upper_out_of_bounds = np.array([5, 20, 30])
        np.testing.assert_array_almost_equal(
            self.bounds.project(input_with_upper_out_of_bounds),
            expected_output_with_upper_out_of_bounds,
        )

    def test_both_out_of_bounds(self):
        input_with_both_bounds_out_of_bounds = [0, 25, 40]
        expected_output_with_both_bounds_out_of_bounds = np.array([1, 20, 30])
        np.testing.assert_array_almost_equal(
            self.bounds.project(input_with_both_bounds_out_of_bounds),
            expected_output_with_both_bounds_out_of_bounds,
        )

    def test_incompatible(self):
        input_with_incompatible_size = [1, 2, 3, 4]
        with self.assertRaises(OptimizationError):
            self.bounds.project(input_with_incompatible_size)

    def test_intersect(self):
        """
        Test the intersect method of the Bounds class.
        """
        # Create Bounds objects for testing
        bounds1 = Bounds([(1, 5), (2, 6), (3, 7)])
        bounds2 = Bounds([(3, 8), (4, 9), (5, 10)])

        # Test intersect with compatible dimensions
        expected_output = Bounds([(3, 5), (4, 6), (5, 7)])
        self.assertEqual(bounds1.intersect(bounds2).bounds, expected_output.bounds)

        # Test intersect with incompatible dimensions
        bounds3 = Bounds([(1, 5), (2, 6), (3, 7), (4, 8)])
        with self.assertRaises(OptimizationError):
            bounds1.intersect(bounds3)

    def test_intersection_with_trust_region(self):
        """
        Test the intersection_with_trust_region method of the Bounds class.
        """
        # Create Bounds object for testing
        bounds = Bounds([(0, 5), (-2, 2), (1, 10), (-6, -5)])

        # Test intersection with compatible dimensions
        center = np.array([2.5, 1.5, 1, -5.5])
        delta = 1.5
        expected_output = Bounds([(1, 4), (0, 2), (1, 2.5), (-6, -5)])
        self.assertEqual(
            bounds.intersection_with_trust_region(center, delta).bounds,
            expected_output.bounds,
        )

        # Test intersection with incompatible dimensions
        center = np.array([2, 0])
        delta = 1.5
        with self.assertRaises(OptimizationError):
            bounds.intersection_with_trust_region(center, delta)

    def test_bounds_for_trust_region_subproblem(self):
        """ """
        # Create Bounds object for testing
        bounds = Bounds([(0, 5), (-2, 2), (1, 10), (-6, -5)])

        # Test intersection with compatible dimensions
        center = np.array([2.5, 1.5, 1, -5.5])
        delta = 1.5
        expected_output = Bounds([(-1.5, 1.5), (-1.5, 0.5), (0, 1.5), (-0.5, 0.5)])
        self.assertEqual(
            bounds.get_bounds_for_trust_region_subproblem(center, delta).bounds,
            expected_output.bounds,
        )

        # Test intersection with incompatible dimensions
        center = np.array([2, 0])
        delta = 1.5
        with self.assertRaises(OptimizationError):
            bounds.intersection_with_trust_region(center, delta)

    def test_subspace(self):
        """
        Test the subspace method of the Bounds class.
        """
        # Create Bounds object for testing
        bounds = Bounds([(0, 5), (-2, 2), (1, 10)])

        # Test subspace with selected unsorted_set_of_variables
        selected_variables = [True, False, True]
        expected_output = Bounds([(0, 5), (1, 10)])
        self.assertEqual(
            bounds.subspace(selected_variables).bounds, expected_output.bounds
        )

        # Test subspace with incompatible dimensions
        selected_variables = [True, False]
        with self.assertRaises(OptimizationError):
            bounds.subspace(selected_variables)

    def test_feasible(self):
        """
        Test the feasible method of the Bounds class.
        """
        # Create Bounds object for testing
        bounds = Bounds([(0, 5), (-2, 2), (1, 10)])

        # Test feasible with a feasible point
        point = np.array([2, 0, 5])
        self.assertTrue(bounds.feasible(point))

        # Test feasible with a point violating the lower bound
        point = np.array([-1, 0, 5])
        self.assertFalse(bounds.feasible(point))

        # Test feasible with a point violating the upper bound
        point = np.array([2, 0, 11])
        self.assertFalse(bounds.feasible(point))

        # Test feasible with a point violating both lower and upper bounds
        point = np.array([-1, 0, 11])
        self.assertFalse(bounds.feasible(point))

        # Test feasible with incompatible dimensions
        point = np.array([2, 0])
        with self.assertRaises(OptimizationError):
            bounds.feasible(point)

    def test_maximum_step_feasible_point(self):
        bounds = Bounds([(1, 2), (1, 2)])
        x = np.array([1.5, 1.5])
        d = np.array([1, 1])
        expected_alpha = 0.5
        expected_indices = np.array([0, 1])

        alpha, indices = bounds.maximum_step(x, d)

        self.assertAlmostEqual(alpha, expected_alpha)
        np.testing.assert_array_equal(indices, expected_indices)

    def test_maximum_step_infeasible_point(self):
        bounds = Bounds([(1, 2), (1, 2)])
        x = np.array([0.5, 0.5])
        d = np.array([1, 1])

        with self.assertRaises(OptimizationError):
            bounds.maximum_step(x, d)

    def test_maximum_step_positive_direction(self):
        bounds = Bounds([(1, 2), (1, 2)])
        x = np.array([1.5, 1.5])
        d = np.array([1, 1])
        expected_alpha = 0.5
        expected_indices = np.array([0, 1])

        alpha, indices = bounds.maximum_step(x, d)

        self.assertAlmostEqual(alpha, expected_alpha)
        np.testing.assert_array_equal(indices, expected_indices)

    def test_maximum_step_negative_direction(self):
        bounds = Bounds([(1, 2), (1, 2)])
        x = np.array([1.5, 1.5])
        d = np.array([-1, -1])
        expected_alpha = 0.5
        expected_indices = np.array([0, 1])

        alpha, indices = bounds.maximum_step(x, d)

        self.assertAlmostEqual(alpha, expected_alpha)
        np.testing.assert_array_equal(indices, expected_indices)

    def test_maximum_step_mixed_direction(self):
        bounds = Bounds([(1, 2), (1, 2)])
        x = np.array([1.5, 1.5])
        d = np.array([1, -1])
        expected_alpha = 0.5
        expected_indices = np.array([0, 1])

        alpha, indices = bounds.maximum_step(x, d)

        self.assertAlmostEqual(alpha, expected_alpha)
        np.testing.assert_array_equal(indices, expected_indices)

    def test_activity_inactive_bounds(self):
        # Test when no bounds are active
        bounds = Bounds([(-1, 3), (0, 1), (2, 4)])
        point = np.array([0, 0.5, 3])
        activity = bounds.activity(point)
        expected_activity = np.array([0, 0, 0])
        np.testing.assert_array_equal(activity, expected_activity)

    def test_activity_lower_bound_active(self):
        # Test when lower bounds are active
        bounds = Bounds([(-1, 3), (0, 1), (2, 4)])
        point = np.array([-1, 0.5, 2])
        activity = bounds.activity(point)
        expected_activity = np.array([-1, 0, -1])
        np.testing.assert_array_equal(activity, expected_activity)

    def test_activity_upper_bound_active(self):
        # Test when upper bounds are active
        bounds = Bounds([(-1, 3), (0, 1), (2, 4)])
        point = np.array([2, 0.5, 4])
        activity = bounds.activity(point)
        expected_activity = np.array([0, 0, 1])
        np.testing.assert_array_equal(activity, expected_activity)

    def test_activity_both_bounds_active(self):
        # Test when both lower and upper bounds are active
        bounds = Bounds([(-1, 3), (0, 1), (2, 4)])
        point = np.array([-1, 0, 4])
        activity = bounds.activity(point)
        expected_activity = np.array([-1, -1, 1])
        np.testing.assert_array_equal(activity, expected_activity)

    def test_activity_feasible_point(self):
        # Test with a feasible point
        bounds = Bounds([(-1, 3), (0, 1), (2, 4)])
        point = np.array([1, 0.5, 3])
        activity = bounds.activity(point)
        expected_activity = np.array([0, 0, 0])
        np.testing.assert_array_equal(activity, expected_activity)

    def test_activity_invalid_point_size(self):
        # Test with an invalid point size
        bounds = Bounds([(-1, 3), (0, 1), (2, 4)])
        point = np.array([1, 2])  # Size does not match the dimension of bounds
        with self.assertRaises(OptimizationError):
            bounds.activity(point)

    def test_activity_infeasible_point(self):
        # Test with an infeasible point
        bounds = Bounds([(-1, 3), (0, 1), (2, 4)])
        point = np.array([4, 0.5, 3])
        with self.assertRaises(OptimizationError):
            bounds.activity(point)

    def test_active_constraints_all_active(self):
        # Test case: All unsorted_set_of_variables are inactive
        bounds = Bounds([(0, 1), (2, 3), (-1, 1)])
        point = np.array([0.5, 2.5, 0])
        expected_status = set()
        active_constraints = bounds.active_constraints(point)
        self.assertSetEqual(active_constraints, expected_status)

    def test_active_constraints_lower_bound_active(self):
        # Test case: Variable 1 is lower bound active
        bounds = Bounds([(0, 1), (2, 3), (-1, 1)])
        point = np.array([0, 2.5, 0])
        expected_status = {0}
        active_constraints = bounds.active_constraints(point)
        self.assertSetEqual(active_constraints, expected_status)

    def test_active_constraints_upper_bound_active(self):
        # Test case: Variable 2 is upper bound active
        bounds = Bounds([(0, 1), (2, 3), (-1, 1)])
        point = np.array([0.5, 3, 0])
        expected_status = {1}
        active_constraints = bounds.active_constraints(point)
        self.assertSetEqual(active_constraints, expected_status)

    def test_active_constraints_both_bounds_active(self):
        # Test case: Variable 1 is lower bound active and Variable 3 is upper bound active
        bounds = Bounds([(0, 1), (2, 3), (-1, 1)])
        point = np.array([0, 2.5, 1])
        expected_status = {0, 2}
        active_constraints = bounds.active_constraints(point)
        self.assertSetEqual(active_constraints, expected_status)

    def test_active_constraints_no_active(self):
        # Test case: All unsorted_set_of_variables are active
        bounds = Bounds([(0, 1), (2, 3), (-1, 1)])
        point = np.array([0, 3, -1])
        expected_status = {0, 1, 2}
        active_constraints = bounds.active_constraints(point)
        self.assertSetEqual(active_constraints, expected_status)

    def test_inactive_constraints_all_inactive(self):
        # Test case: All unsorted_set_of_variables are inactive
        bounds = Bounds([(0, 1), (2, 3), (-1, 1)])
        point = np.array([0.5, 2.5, 0])
        expected_status = {0, 1, 2}
        active_constraints = bounds.inactive_constraints(point)
        self.assertSetEqual(active_constraints, expected_status)

    def test_inactive_constraints_lower_bound_inactive(self):
        # Test case: Variable 1 is lower bound active
        bounds = Bounds([(0, 1), (2, 3), (-1, 1)])
        point = np.array([0, 2.5, 0])
        expected_status = {1, 2}
        active_constraints = bounds.inactive_constraints(point)
        self.assertSetEqual(active_constraints, expected_status)

    def test_inactive_constraints_upper_bound_inactive(self):
        # Test case: Variable 2 is upper bound active
        bounds = Bounds([(0, 1), (2, 3), (-1, 1)])
        point = np.array([0.5, 3, 0])
        expected_status = {0, 2}
        active_constraints = bounds.inactive_constraints(point)
        self.assertSetEqual(active_constraints, expected_status)

    def test_inactive_constraints_both_bounds_inactive(self):
        # Test case: Variable 1 is lower bound active and Variable 3 is upper bound active
        bounds = Bounds([(0, 1), (2, 3), (-1, 1)])
        point = np.array([0, 2.5, 1])
        expected_status = {1}
        active_constraints = bounds.inactive_constraints(point)
        self.assertSetEqual(active_constraints, expected_status)

    def test_inactive_constraints_no_inactive(self):
        # Test case: All unsorted_set_of_variables are active
        bounds = Bounds([(0, 1), (2, 3), (-1, 1)])
        point = np.array([0, 3, -1])
        expected_status = set()
        active_constraints = bounds.inactive_constraints(point)
        self.assertSetEqual(active_constraints, expected_status)

    def test_breakpoints(self):
        # Test with a feasible point and direction
        bounds_list = [(0, 1), (-1, 2), (1, 3)]
        bounds = Bounds(bounds_list)
        point = np.array([0.5, 0, 2])
        direction = np.array([1, 0.5, -1])
        result = bounds.breakpoints(point, direction)
        expected_result = [(0, 0.5), (2, 1), (1, 4)]
        self.assertEqual(result, expected_result)

    def test_breakpoints_invalid_direction(self):
        # Test with an invalid direction size
        bounds_list = [(0, 1), (-1, 2), (1, 3)]
        bounds = Bounds(bounds_list)
        point = np.array([0, 0, 0])
        direction = np.array([1, 0.5])  # Size does not match the dimension of bounds
        with self.assertRaises(OptimizationError):
            bounds.breakpoints(point, direction)

    def test_breakpoints_infeasible_point(self):
        # Test with an infeasible point
        bounds_list = [(0, 1), (-1, 2), (1, 3)]
        bounds = Bounds(bounds_list)
        point = np.array([0, 2, 4])
        direction = np.array([1, 0.5, -1])
        with self.assertRaises(OptimizationError):
            bounds.breakpoints(point, direction)

    def test_projected_direction(self):
        # Define the bounds for the unsorted_set_of_variables
        bounds = Bounds([(-1, 1), (-1, 1), (-1, 1)])

        # Test case 1: x_current is inactive
        x_current = np.array([0.0, 0.0, 0.0])
        direction = np.array([-1.0, 2.0, 0.0])
        expected_direction = direction
        projected_direction = bounds.projected_direction(x_current, direction)

        np.testing.assert_array_almost_equal(projected_direction, expected_direction)

        # Test case 2: x_current is active
        x_current = np.array([-1.0, 0.0, 0.0])
        direction = np.array([-0.5, -1.0, 0.0])
        expected_direction = direction
        expected_direction[0] = 0
        projected_direction = bounds.projected_direction(x_current, direction)

        np.testing.assert_array_almost_equal(projected_direction, expected_direction)

        # Test case 3: x_current is partially active
        x_current = np.array([1.0, 1.0, 0.0])
        g_current = np.array([-1.0, 1.0, 5.0])
        expected_direction = direction
        expected_direction[1] = 0

        projected_direction = bounds.projected_direction(x_current, direction)

        np.testing.assert_array_almost_equal(projected_direction, expected_direction)


import unittest
import numpy as np


class TestBounds(unittest.TestCase):
    def test_generalized_cauchy_point_convergence(self):
        # Test case inputs
        bounds = Bounds([(-100, 100), (-100, 100)])

        iterate = np.array([0, 0])
        gradient = np.array([1, 1])
        hessian = np.array([[1, 0], [0, 1]])

        # Call the generalized_cauchy_point function
        result = bounds.generalized_cauchy_point(iterate, gradient, hessian)

        # Define the expected result
        expected_result = np.array([-1, -1])

        # Perform the assertion
        np.testing.assert_array_almost_equal(result, expected_result)

    def test_generalized_cauchy_point_active_bounds(self):
        # Test case inputs
        bounds = Bounds([(-0.5, 100), (-100, 100)])

        iterate = np.array([0, 0])
        gradient = np.array([1, 1])
        hessian = np.array([[1, 0], [0, 1]])

        # Call the generalized_cauchy_point function
        result = bounds.generalized_cauchy_point(iterate, gradient, hessian)

        # Define the expected result
        expected_result = np.array([-0.5, -1])

        # Perform the assertion
        np.testing.assert_array_almost_equal(result, expected_result)

    def test_generalized_cauchy_point_large(self):

        iterate = np.array(
            [
                -1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
            ]
        )
        gradient = np.array(
            [
                -0.0,
                -0.0,
                -0.0,
                -0.0,
                -0.0,
                -0.0,
                -5898.0,
                -4947.0,
                -6562.0,
                -5374.0,
                -6084.0,
                -5497.0,
                -13291.0,
                -19981.0,
                -14428.0,
                -24615.0,
                -17003.0,
                -21160.0,
                -18994.0,
                -3095.0,
                -7788.0,
                -1120.0,
                -4629.0,
                -12404.0,
                -17973.0,
                -9194.0,
                -12342.0,
                -9048.0,
            ]
        )
        hessian = np.array(
            [
                [
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                ],
                [
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                ],
            ]
        )
        lower_bounds = np.array(
            [
                -101.0,
                -99.0,
                -99.0,
                -99.0,
                -99.0,
                -99.0,
                -100.0,
                -100.0,
                -100.0,
                -100.0,
                -100.0,
                -100.0,
                -99.0,
                -99.0,
                -99.0,
                -99.0,
                -99.0,
                -99.0,
                -99.0,
                -100.0,
                -100.0,
                -100.0,
                -100.0,
                -100.0,
                -100.0,
                -100.0,
                -100.0,
                -100.0,
            ]
        )
        upper_bounds = np.array(
            [
                99.0,
                101.0,
                101.0,
                101.0,
                101.0,
                101.0,
                100.0,
                100.0,
                100.0,
                100.0,
                100.0,
                100.0,
                101.0,
                101.0,
                101.0,
                101.0,
                101.0,
                101.0,
                101.0,
                100.0,
                100.0,
                100.0,
                100.0,
                100.0,
                100.0,
                100.0,
                100.0,
                100.0,
            ]
        )

        bounds = Bounds.from_bounds(lower_bounds, upper_bounds)

        # Call the generalized_cauchy_point function
        result = bounds.generalized_cauchy_point(iterate, gradient, hessian)

        # Define the expected result
        expected_result = np.array(
            [
                -1,
                1,
                1,
                1,
                1,
                1,
                100,
                100,
                100,
                100,
                100,
                100,
                101,
                101,
                101,
                101,
                101,
                101,
                101,
                100,
                100,
                100,
                100,
                100,
                100,
                100,
                100,
                100,
            ]
        )

        np.testing.assert_array_almost_equal(result, expected_result)

    def test_generalized_cauchy_point_2(self):
        # Test case inputs
        bounds = Bounds([(-1, 1), (-1, 1)])

        iterate = np.array([0, 0])
        gradient = np.array([10, 10])
        hessian = np.array([[1, 0], [0, 1]])

        # Call the generalized_cauchy_point function
        result = bounds.generalized_cauchy_point(iterate, gradient, hessian)

        # Define the expected result
        expected_result = np.array([-1, -1])

        # Perform the assertion
        np.testing.assert_array_almost_equal(result, expected_result)

    def test_generalized_cauchy_point_all_active_bounds(self):
        # Test case inputs
        bounds = Bounds([(-0.5, 100), (-0.4, 100)])

        iterate = np.array([0, 0])
        gradient = np.array([1, 1])
        hessian = np.array([[1, 0], [0, 1]])

        # Call the generalized_cauchy_point function
        result = bounds.generalized_cauchy_point(iterate, gradient, hessian)

        # Define the expected result
        expected_result = np.array([-0.5, -0.4])

        # Perform the assertion
        np.testing.assert_array_almost_equal(result, expected_result)

    def test_conjugate_gradient_convergence(self):

        bounds = Bounds([(None, None), (None, None)])

        gradient = np.array([1, 1])
        hessian = np.array([[1, 0], [0, 1]])
        expected_solution = np.array([-1, -1])
        expected_diagnostic = ConjugateGradientDiagnostic.CONVERGENCE
        solution, diagnostic = bounds.truncated_conjugate_gradient(
            gradient=gradient,
            hessian=hessian,
        )
        np.testing.assert_array_almost_equal(solution, expected_solution)
        self.assertEqual(diagnostic, expected_diagnostic)

        gradient = np.array([1, 1])
        hessian = np.array([[1, 0], [0, 10]])
        expected_solution = np.array([-1, -0.1])
        expected_diagnostic = ConjugateGradientDiagnostic.CONVERGENCE
        solution, diagnostic = bounds.truncated_conjugate_gradient(
            gradient=gradient,
            hessian=hessian,
        )
        np.testing.assert_array_almost_equal(solution, expected_solution)
        self.assertEqual(diagnostic, expected_diagnostic)

    def test_conjugate_gradient_convergence_negative_curvature(self):

        bounds = Bounds([(-100, 100), (-100, 100)])

        gradient = np.array([1, 1])
        hessian = np.array([[1, 0], [0, -1]])
        expected_solution = np.array([-100, -100])
        expected_diagnostic = ConjugateGradientDiagnostic.NEGATIVE_CURVATURE
        solution, diagnostic = bounds.truncated_conjugate_gradient(
            gradient=gradient,
            hessian=hessian,
        )
        np.testing.assert_array_almost_equal(solution, expected_solution)
        self.assertEqual(diagnostic, expected_diagnostic)

    def test_conjugate_gradient_convergence_out_of_trust_region(self):

        bounds = Bounds([(None, None), (-0.05, None)])

        gradient = np.array([1, 1])
        hessian = np.array([[1, 0], [0, 1]])
        expected_solution = np.array([-0.05, -0.05])
        expected_diagnostic = ConjugateGradientDiagnostic.OUT_OF_TRUST_REGION
        solution, diagnostic = bounds.truncated_conjugate_gradient(
            gradient=gradient,
            hessian=hessian,
        )
        np.testing.assert_array_almost_equal(solution, expected_solution)
        self.assertEqual(diagnostic, expected_diagnostic)

        gradient = np.array([1, 1])
        hessian = np.array([[1, 0], [0, 10]])
        expected_solution = np.array([-0.05, -0.05])
        expected_diagnostic = ConjugateGradientDiagnostic.OUT_OF_TRUST_REGION
        solution, diagnostic = bounds.truncated_conjugate_gradient(
            gradient=gradient,
            hessian=hessian,
        )
        np.testing.assert_array_almost_equal(solution, expected_solution)
        self.assertEqual(diagnostic, expected_diagnostic)

    def test_conjugate_gradient_subspace_convergence(self):

        bounds = Bounds([(None, None), (None, None)])

        iterate = np.array([0, 0])
        gradient = np.array([1, 1])
        hessian = np.array([[1, 0], [0, 1]])
        radius = 100
        expected_solution = np.array([-1, -1])
        expected_diagnostic = ConjugateGradientDiagnostic.CONVERGENCE
        solution, diagnostic = bounds.truncated_conjugate_gradient_subspace(
            iterate=iterate,
            gradient=gradient,
            hessian=hessian,
            radius=radius,
        )
        np.testing.assert_array_almost_equal(solution, expected_solution)
        self.assertEqual(diagnostic, expected_diagnostic)

        iterate = np.array([1, 1])
        gradient = np.array([1, 1])
        hessian = np.array([[1, 0], [0, 1]])
        radius = 100
        expected_solution = np.array([0, 0])
        expected_diagnostic = ConjugateGradientDiagnostic.CONVERGENCE
        solution, diagnostic = bounds.truncated_conjugate_gradient_subspace(
            iterate=iterate,
            gradient=gradient,
            hessian=hessian,
            radius=radius,
        )
        np.testing.assert_array_almost_equal(solution, expected_solution)
        self.assertEqual(diagnostic, expected_diagnostic)

    def test_conjugate_gradient_subspace_convergence(self):

        bounds = Bounds([(None, None), (None, None)])

        iterate = np.array([0, 0])
        gradient = np.array([1, 1])
        hessian = np.array([[1, 0], [0, 1]])
        radius = 100
        expected_solution = np.array([-1, -1])
        expected_diagnostic = ConjugateGradientDiagnostic.CONVERGENCE
        solution, diagnostic = bounds.truncated_conjugate_gradient_subspace(
            iterate=iterate,
            gradient=gradient,
            hessian=hessian,
            radius=radius,
        )
        np.testing.assert_array_almost_equal(solution, expected_solution)
        self.assertEqual(diagnostic, expected_diagnostic)

        iterate = np.array([1, 1])
        gradient = np.array([1, 1])
        hessian = np.array([[1, 0], [0, 1]])
        radius = 100
        expected_solution = np.array([0, 0])
        expected_diagnostic = ConjugateGradientDiagnostic.CONVERGENCE
        solution, diagnostic = bounds.truncated_conjugate_gradient_subspace(
            iterate=iterate,
            gradient=gradient,
            hessian=hessian,
            radius=radius,
        )
        np.testing.assert_array_almost_equal(solution, expected_solution)
        self.assertEqual(diagnostic, expected_diagnostic)

    def test_conjugate_gradient_subspace_active(self):

        bounds = Bounds([(-0.5, None), (None, None)])

        iterate = np.array([0, 0])
        gradient = np.array([1, 1])
        hessian = np.array([[1, 0], [0, 1]])
        radius = 100
        expected_solution = np.array([-0.5, -1])
        expected_diagnostic = ConjugateGradientDiagnostic.CONVERGENCE
        solution, diagnostic = bounds.truncated_conjugate_gradient_subspace(
            iterate=iterate,
            gradient=gradient,
            hessian=hessian,
            radius=radius,
        )
        np.testing.assert_array_almost_equal(solution, expected_solution)
        self.assertEqual(diagnostic, expected_diagnostic)

        bounds = Bounds([(0.5, None), (None, None)])

        iterate = np.array([1, 1])
        gradient = np.array([1, 1])
        hessian = np.array([[1, 0], [0, 1]])
        radius = 100
        expected_solution = np.array([0.5, 0])
        expected_diagnostic = ConjugateGradientDiagnostic.CONVERGENCE
        solution, diagnostic = bounds.truncated_conjugate_gradient_subspace(
            iterate=iterate,
            gradient=gradient,
            hessian=hessian,
            radius=radius,
        )
        np.testing.assert_array_almost_equal(solution, expected_solution)
        self.assertEqual(diagnostic, expected_diagnostic)


if __name__ == '__main__':
    unittest.main()



"""File bfgs.py

:author: Michel Bierlaire
:date: Sat Jun 24 17:34:20 2023

Functions for the BFGS algorithm
"""

import logging

import numpy as np

from biogeme_optimization.exceptions import OptimizationError

logger = logging.getLogger(__name__)


def bfgs(hessian_approx: np.ndarray, delta_x: np.ndarray, delta_g: np.ndarray):
    """Update the BFGS matrix. Formula (13.12) of `Bierlaire (2015)`_
    where the method proposed by `Powell (1977)`_ is applied

    .. _`Bierlaire (2015)`: http://optimizationprinciplesalgorithms.com/
    .. _`Powell (1977)`: https://link.springer.com/content/pdf/10.1007/BFb0067703.pdf

    :param hessian_approx: current approximation of the Hessian
    :type hessian_approx: numpy.array (2D)

    :param delta_x: difference between two consecutive iterates.
    :type delta_x: numpy.array (1D)

    :param delta_g: difference between two consecutive gradients.
    :type delta_g: numpy.array (1D)

    :return: updated approximation of the inverse of the Hessian.
    :rtype: numpy.array (2D)

    """
    if np.all(delta_x == 0):
        raise OptimizationError('In BFGS, the step cannot be zero')

    if np.all(delta_g == 0):
        raise OptimizationError('In BFGS, the gradient difference cannot be zero')
    h_d = hessian_approx @ delta_x
    d_h_d = np.inner(delta_x, h_d)
    denominator = np.inner(delta_x, delta_g)
    if denominator >= 0.2 * d_h_d:
        eta = delta_g
    else:
        theta = 0.8 * d_h_d / (d_h_d - denominator)
        eta = theta * delta_g + (1 - theta) * h_d

    delta_x_eta = np.inner(delta_x, eta)
    hessian_update = np.outer(eta, eta) / delta_x_eta - np.outer(h_d, h_d) / d_h_d
    return hessian_approx + hessian_update


def inverse_bfgs(
    inverse_hessian_approx: np.ndarray, delta_x: np.ndarray, delta_g: np.ndarray
):
    """Update the inverse BFGS matrix. Formula (13.13) of `Bierlaire (2015)`_

    .. _`Bierlaire (2015)`: http://optimizationprinciplesalgorithms.com/

    :param inverse_hessian_approx: current approximation of the inverse of the Hessian
    :type inverse_hessian_approx: numpy.array (2D)

    :param delta_x: difference between two consecutive iterates.
    :type delta_x: numpy.array (1D)

    :param delta_g: difference between two consecutive gradients.
    :type delta_g: numpy.array (1D)

    :return: updated approximation of the inverse of the Hessian.
    :rtype: numpy.array (2D)
    """
    dimension = len(delta_x)

    denominator = np.inner(delta_x, delta_g)
    if denominator <= 0.0:
        raise OptimizationError(
            f"Unable to perform inverse BFGS update as d'y = {denominator} <= 0"
        )
    d_y = np.outer(delta_x, delta_g)
    y_d = np.outer(delta_g, delta_x)
    d_d = np.outer(delta_x, delta_x)
    eye = np.identity(dimension)
    return (
        (eye - (d_y / denominator))
        @ inverse_hessian_approx
        @ (eye - (y_d / denominator))
    ) + d_d / denominator



"""File test_hybrid_function.py

:author: Michel Bierlaire
:date: Sun Jul  2 16:14:20 2023

Tests for the hybrid_function module
"""

import unittest

import numpy as np

from biogeme_optimization.bounds import Bounds
from biogeme_optimization.function import FunctionData
from biogeme_optimization.hybrid_function import HybridFunction
from examples import MyFunctionToMinimize


class TestHybridFunction(unittest.TestCase):
    def setUp(self):
        # Create an instance of MyFunctionToMinimize with dimension 3
        self.dimension = 3
        self.function = MyFunctionToMinimize(self.dimension)
        self.proportion = 0.5
        self.bounds = Bounds(
            [
                (None, None),
                (None, None),
                (None, None),
            ]
        )
        self.hybrid_func = HybridFunction(self.function, self.proportion, self.bounds)

    def test_calculate_function_and_derivatives_analytical(self):
        # Test when analytical Hessian is used
        iterate = np.array([1, 2, 3])
        function_data = self.hybrid_func._calculate_function_and_derivatives_analytical(
            iterate
        )
        expected_function_data = FunctionData(
            function=7,
            gradient=iterate,
            hessian=np.eye(self.dimension),
        )
        self.assertEqual(function_data.function, expected_function_data.function)
        np.testing.assert_array_almost_equal(
            function_data.gradient, expected_function_data.gradient
        )
        np.testing.assert_array_almost_equal(
            function_data.hessian, expected_function_data.hessian
        )
        self.assertEqual(self.hybrid_func.number_of_analytical_hessians, 1)
        self.assertEqual(self.hybrid_func.number_of_matrices, 1)
        # Add more assertions for the specific behavior and expected values

    def test_calculate_function_and_derivatives_bfgs(self):
        # Test when BFGS approximation is used
        iterate = np.array([1, 2, 3])
        self.function.x = iterate
        function_data = self.hybrid_func._calculate_function_and_derivatives_bfgs(
            iterate
        )
        expected_function_data = FunctionData(
            function=7,
            gradient=iterate,
            hessian=np.eye(self.dimension),
        )
        self.assertEqual(function_data.function, expected_function_data.function)
        np.testing.assert_array_almost_equal(
            function_data.gradient, expected_function_data.gradient
        )
        np.testing.assert_array_almost_equal(
            function_data.hessian, expected_function_data.hessian
        )
        self.assertEqual(self.hybrid_func.number_of_matrices, 1)

        iterate = np.array([0, 0, 0])
        self.function.x = iterate
        function_data = self.hybrid_func._calculate_function_and_derivatives_bfgs(
            iterate
        )
        expected_function_data = FunctionData(
            function=0,
            gradient=iterate,
            hessian=np.eye(self.dimension),
        )
        # The solution is optimal
        self.assertIsNone(function_data)

    def test_calculate_function_and_derivatives(self):
        # Test when analytical Hessian is used
        iterate = np.array([1, 2, 3])
        self.function.x = iterate
        self.hybrid_func.calculate_function_and_derivatives(iterate)
        self.assertEqual(self.hybrid_func.number_of_analytical_hessians, 1)
        self.assertEqual(self.hybrid_func.number_of_matrices, 1)
        # Add more assertions for the specific behavior and expected values

        # Test when BFGS approximation is used
        iterate = np.array([4, 5, 6])
        self.function.x = iterate
        self.hybrid_func.calculate_function_and_derivatives(iterate)
        self.assertEqual(self.hybrid_func.number_of_matrices, 2)
        # Add more assertions for the specific behavior and expected values

    def test_message(self):
        # Test when there are no numerical issues
        self.hybrid_func.number_of_analytical_hessians = 2
        self.hybrid_func.number_of_matrices = 5
        self.assertEqual(self.hybrid_func.message(), '2/5 = 40.0%')

        # Test when there are numerical issues
        self.hybrid_func.number_of_numerical_issues = 1
        self.assertEqual(self.hybrid_func.message(), '2/5 = 40.0% [1 numerical issues]')

        # Test when there are no matrices involved
        self.hybrid_func.number_of_matrices = 0
        self.assertEqual(self.hybrid_func.message(), '')


if __name__ == '__main__':
    unittest.main()
